# -*- coding: utf_8 -*-

from panda3d.core import *
from direct.showbase.DirectObject import DirectObject
import direct.directbase.DirectStart
#from direct.filter.CommonFilters import CommonFilters
import random


class SkySphere(DirectObject): 

    def __init__(self, size, tex, transparent = False):       
        self.sphere = loader.loadModel("InvertedSphere.egg") 
        self.sphere.setTexGen(TextureStage.getDefault(), TexGenAttrib.MWorldPosition) 
        self.sphere.setTexProjector(TextureStage.getDefault(), render, self.sphere) 
        self.sphere.setTexPos(TextureStage.getDefault(), 0, 0, 0) 
        self.sphere.setTexScale(TextureStage.getDefault(), .5) 
        tex = loader.loadCubeMap(tex + "_#.png", okMissing = True)
        self.sphere.setTexture(tex) 
        self.sphere.setLightOff() 
        self.sphere.setScale(size) 
        self.sphere.setBin('background', 1) 
        self.sphere.setDepthWrite(0) 
        self.sphere.setTransparency(TransparencyAttrib.MAlpha)
        self.sphere.reparentTo(render) 

    def model(self):
        return self.sphere;


class AField:
    
    def __init__(self, player):
        self.player = player
        self.sec_count = 5000 
        self.field_size = (4000, 4000, 4000)
        self.density = 250
        self.sec_size = 300
        self.sec_ast_count = (50, 200)
        self.sec_template = [1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 3]
        self.sec_list = {}
        self.sec_cache = []
        self.prepare_cache(20)
        self.static_root = render.attachNewNode('static_root')
        self.dynamic_root = render.attachNewNode('dynamic_root')
        
    def add_sector(self, pos):
        key = ' '.join(map(str,pos))
        if not key in self.sec_list.keys():
            cache_id = random.randrange(len(self.sec_cache))
            sector = self.sec_cache[cache_id].get_sector_data_copy()
            sector.pos = Vec3(*pos)
            sector.cache_id = cache_id
            self.sec_list[key] = sector
            #self.sec_list[pos].show()
    
    def prepare_cache(self, count):
        for i in xrange(count):
            sector = ASector(Vec3(0,0,0), 
                             template = self.sec_template,
                             size = (self.sec_size, self.sec_size, self.sec_size),
                             count = self.sec_ast_count,
                             density = self.density)
            sector.fill()
            sector.make_static_geom()
            sector.make_dynamic_geom()
            self.sec_cache.append(sector)
    
    def generate(self):
        dx = self.field_size[0]/self.density
        dy = self.field_size[1]/self.density 
        dz = self.field_size[2]/self.density
        for i in xrange(self.sec_count):
            x = int(random.random()*dx - dx*0.5) 
            y = int(random.random()*dy - dy*0.5)
            z = int(random.random()*dz - dz*0.5)
            self.add_sector((x, y, z))
            
    def stat_info(self):
        sectors = 0
        asteroids = 0
        for s in self.sec_list.values():
            sectors += 1
            asteroids += len(s.asteroids)
        print 'SEC: %i\nAST:%i' % (sectors, asteroids)
    
    def update_generic(self, task = None):
        for p, s in self.sec_list.items():
            pass
        if task:
            return task.cont
            
    def update_sectors(self, task = None):
        active, inactive, invisible = 0, 0, 0
        dist_sq_dyn = (self.density ** 2)# + (self.density * 0.5)
        dist_sq_clr = 800 ** 2
        for sec in self.sec_list.values():
            sec_sq_dist = (self.player.getPos(render) - (sec.pos * self.density)).length_squared()
            if sec_sq_dist < dist_sq_dyn:
                sec.set_dynamic(self.sec_cache, self.dynamic_root)
                active += 1
            elif sec_sq_dist < dist_sq_clr:
                sec.set_static(self.sec_cache, self.static_root)
                inactive += 1
            else:
                sec.clear_geom()
                invisible += 1
        if task:
            return task.again
        else:
            print 'Act:%i\nInact:%i\nInvis:%i' % (active, inactive, invisible)
            print self.player.getPos(render)


class ASector:
    
    def __init__(self, pos, count = (25,30), size = (100, 100, 100), 
                  template = [1, ], density = 50):
        self.pos = pos
        self.template = template
        self.base_count = count
        self.size = size
        self.density = density
        self.count = 0
        #self.polar_template
        self.asteroids = []
        self.active = False
        self.active_list = {}
        self.static = None
        self.dynamic = None
        self.cache_id = -1
        #self._fill()
        
    def fill(self):
        if not self.count:
            count_deg = 0
            self.count = random.randrange(*self.base_count)
            for i in xrange(self.count):
                size = random.sample(self.template, 1)[0]
                pos = Vec3(random.random() * self.size[0] - self.size[0] * 0.5, 
                           random.random() * self.size[1] - self.size[1] * 0.5, 
                           random.random() * self.size[2] - self.size[2] * 0.5)
                hpr = Vec3(random.random() * 360, random.random() * 360, random.random() * 360)
                new_ast = Asteroid(pos, hpr, size)
                valid_pos = True
                for a in self.asteroids:
                    if (new_ast.pos - a.pos).length_squared() < \
                            ((new_ast.scale + a.scale) * 0.9) ** 2:
                        valid_pos = False
                        count_deg += 1
                        break
                if valid_pos:
                    self.asteroids.append(new_ast)
            print self.count, '-', count_deg
            #self.count -= count_deg

    def make_static_geom(self):
        self.static = NodePath('sector_static')
        lod = LODNode('sector_static_lod')
        lod_np = NodePath(lod)
        ranges = ((400, 0), (800, 400), (1200, 800))
        sizes = (0, 1, 2)
        names = ('near', 'mid', 'far')
        models = {1:'asteroid_s', 2:'asteroid_m', 3:'asteroid'}
        for i in range(3):
            root = NodePath('sector_static_' + names[i])
            for a in self.asteroids:
                if a.size > sizes[i]:
                    #mdl = loader.loadModel('aster_far').find('**/+GeomNode')
                    mdl = loader.loadModel(models[a.size-i]).find('**/+GeomNode')
                    mdl.reparentTo(root)
                    mdl.setPosHpr(a.pos, a.hpr)
                    mdl.setScale(a.scale)
            root.flattenStrong()
            lod.addSwitch(*ranges[i])
            root.reparentTo(lod_np)
        
        #root.setPos(self.pos[0] * 0.5 * self.size[0], 
        #            self.pos[1] * 0.5 * self.size[1], 
        #            self.pos[2] * 0.5 * self.size[2])
        lod_np.reparentTo(self.static)
        #loader.loadModel('zup-axis').reparentTo(self.static)
        
    def make_dynamic_geom(self):
        models = {1:'asteroid_s', 2:'asteroid', 3:'asteroid'}
        self.dynamic = NodePath('sector_dynamic')
        for a in self.asteroids:
            #mdl = loader.loadModel('aster_far')#.find('**/+GeomNode')
            mdl = loader.loadModel(models[a.size])#.find('**/+GeomNode')
            mdl.reparentTo(self.dynamic)
            mdl.setPosHpr(a.pos, a.hpr)
            mdl.setScale(a.scale)
        #loader.loadModel('zup-axis').reparentTo(self.dynamic)
    
    def get_sector_data_copy(self):
        c = ASector(self.pos, self.base_count, self.size, self.template, self.density)
        c.count = self.count
        #for i in xrange(c.count):
        #    pos = Vec3(self.asteroids[i].pos)
        #    hpr = Vec3(self.asteroids[i].hpr)
        #    size = self.asteroids[i].size
        for a in self.asteroids:
            pos = Vec3(a.pos)
            hpr = Vec3(a.hpr)
            size = a.size
            c.asteroids.append(Asteroid(pos, hpr, size))
        return c
    
    def clear_geom(self):
        if self.static:
            self.static.removeNode()
            self.static = None
        if self.dynamic:
            self.dynamic.removeNode()
            self.dynamic = None
        self.active = False
                           
    def set_dynamic(self, cache, root):
        if not self.active and not self.dynamic:
            self.clear_geom()
            if self.cache_id <0:
                self.make_dynamic_geom()
                self.dynamic.reparentTo(root)
            else:
                self.dynamic = cache[self.cache_id].dynamic.copyTo(root)
            self.active = True
            self.dynamic.setPos(self.pos * self.density)
            #print 'DYNAMIC', self.pos
                                
    def set_static(self, cache, root):
        if not self.static:
            self.clear_geom()
            if self.cache_id <0:
                self.make_static_geom()
                self.static.reparentTo(root)
            else:
                self.static = cache[self.cache_id].static.copyTo(root)
            #self.static.show()
            self.static.setPos(self.pos * self.density)



class Asteroid:
    
    def __init__(self, pos, hpr, size):
        scales = {1:(1,2), 2:(4,6), 3:(12,18)}
        self.pos = pos
        self.hpr = hpr
        self.size = size
        self.scale = random.randrange(*scales[self.size])
        self.speed = None
        self.rspeed = None


class Main(DirectObject):
    
    def __init__(self):
        base.setBackgroundColor(0,0,0)
        self.sky = SkySphere(2500, './tex/sky/sky')
        #self.sky2 = SkySphere(1200, './tex/sky2/sky', True)
        #self.sky3 = SkySphere(1500, './tex/sky3/sky', True)
        #self.sky4 = SkySphere(1200, './tex/sky3/sky', True)
        #self.sky4.sphere.setHpr(90, 90, 90)
        print 'prepare field...'
        self.field = AField(base.camera)
        #for i in xrange(100):
            #pos = Vec3(int(random.random()*20), 
            #           int(random.random()*20),
            #           int(random.random()*20))
            #self.field.add_sector(pos)
            #self.field.sec_list[pos].set_static(self.field.sec_cache)
            #self.field.sec_list[pos].set_dynamic(self.field.sec_cache, render)
        print 'generate field...'
        self.field.generate()
        print 'get stat...'
        self.field.stat_info()
        print 'first update...'
        self.field.update_sectors()
        self.setup_light()
        #base.camLens.setFar(1000)
        #self.sky.sphere.clearFog()
        taskMgr.doMethodLater(1.0, self.field.update_sectors, 'update_sectors')
        taskMgr.add(self.update_sky, 'update_sky')
    def update_sky(self, task):
        self.sky.sphere.setPos(base.camera.getPos())
        #self.sky3.sphere.setPos(base.camera.getPos())
        #self.sky4.sphere.setPos(base.camera.getPos())
        return task.cont
        
    def setup_light(self):
        #expfog = Fog("FOG")
        #expfog.setColor(0.01, 0.01, 0.01)
        #expfog.setExpDensity(0.005)
        #self.field.static_root.setFog(expfog)
        
        #dlight = DirectionalLight('dlight')
        #dlight.setColor(Vec4(1, 1, 1, 1))
        #dlight.setShadowCaster(True, 512, 512)
        #dlnp = render.attachNewNode(dlight)
        #dlnp.setHpr(0, -60, 0)
        #self.field.static_root.setLight(dlnp)

        spotlight = camera.attachNewNode( Spotlight( "spotlight" ) )
        spotlight.node().setColor( Vec4( 1, 1, 1, 1 ) )
        spotlight.node().setShadowCaster(True, 512, 512)
        spotlight.node().setLens( PerspectiveLens() )
        spotlight.node().getLens().setFov( 33, 33 )
        spotlight.node().setAttenuation( Vec3( 0.0, 0.0, 0.0001 ) ) 
        spotlight.node().setExponent( 128.0 )
        
        #self.field.dynamic_root.setLight( spotlight )
        #self.field.dynamic_root.setShaderAuto() 
        
        render.setLight( spotlight )        
        render.setShaderAuto() 
        
        #filters = CommonFilters(base.win, base.cam)
        #filters.setBloom(mintrigger = 0.5)

loader.loadModel('box').reparentTo(render)

world = Main()
#render.analyze()
world.field.static_root.analyze()
world.field.dynamic_root.analyze()
#render.ls()
PStatClient.connect()
run()
